(*2*)

let rec insere l (u, v, w) = match l with
    | [] -> [(u, v, w)]
    | (a, b, c) :: q -> if w < c then (u, v, w) :: l
        else (a, b, c) :: (insere q (u, v, w));;


let rec insere_liste l ll i = match ll with
    | [] -> l
    | (a,b) :: q -> insere (insere_liste l q i) (i,a,b);;


(*3*)

let prim g =
  let n = Array.length g in
  let vprime = Array.make n false in
  let arbre = Array.make n [] in
  let f = ref [] in
  vprime.(0) <- true;
  f := insere_liste !f g.(0) 0;
  let nbsommets = ref 1 in
  while !nbsommets < n do
    let ((u, v, w) :: q) = !f in
    f := q;
    if not (vprime.(v)) then
      (arbre.(u) <- (v, w) :: arbre.(u);
      arbre.(v) <- (u,w) :: arbre.(v);
      incr nbsommets; 
        vprime.(v) <- true;
        f := insere_liste !f g.(v) v)
  done;
  arbre;;
 
 
 
let g = [|[(1,2);(3,8);(2,5)];[(0,2);(3,9)];[(0,5);(4,1);(6,4)];[(1,9);(0,8);(5,7)];[(2,1);(6,3);(5,6)];
[(3,7);(4,6)];[(2,4);(4,3)]|];;

prim g;;


(*4*)

let distance (x1,y1) (x2,y2) = abs(x1-x2) + abs(y1 - y2);;

(*5*)

let longueur_chemin points sigma =
  let n = Array.length points in
  let long = ref (distance points.(sigma.(0)) points.(sigma.(n - 1))) in
  for i = 0 to (n - 2) do
    long := !long + (distance points.(sigma.(i)) points.(sigma.(i + 1)))
  done;
  !long;;



let points = [|(-2, 5); (6, -6); (-10, -7); (-1, 0); (5, 2); (7, 1); (7, 7); (-7, -10); (-10,1); (2,7)|];;



let construction_graphe points =
  let n = Array.length points in
  let g = Array.make n [] in
  for i = 0 to (n - 2) do
    for j = i + 1 to (n - 1) do
      g.(i) <- (j, (distance points.(i) points.(j))) :: g.(i);
      g.(j) <- (i, (distance points.(i) points.(j))) :: g.(j)
    done;
  done;
  g;;
  
 let g = construction_graphe points;;

prim g;;

let voyageur points =
  let g = construction_graphe points in
  let gp = prim g in
  let n = Array.length g in
  let sigma = Array.make n 0 in
  let visit = Array.make n false in
  let nb = ref 1 in
  let rec parcours lis = match lis with
      | [] -> () 
      | (s, w) :: q -> if not (visit.(s)) then
          (visit.(s) <- true;
            sigma.(!nb) <- s;
            incr nb;
            parcours gp.(s)); parcours q
  in visit.(0) <- true; parcours gp.(0);
  sigma
;;
 

 
voyageur points;;
longueur_chemin points (voyageur points);;



(* Montrons que l(sigma0) <= 2 min l(sigma) ; On note T la somme des poids des arêtes de l'arbre couvrant de poids minimal.
- on peut voir que l(sigma0) <= 2 T. En effet lors du parcours en profondeur de l'arbre on parcourt toutes les arêtes deux fois 
mais dans l(sigma0) on peut par moment racourcir le chemin grace à l'inégalité triangulaire
- soit sigma la permutation optimale, en considérant l'arbre dont les arêtes sont (v_sigma(0), v_sigma(1)) ; ...
 ; (v_sigma(n-2),v_sigma(n-1)) le poids de cet arbre vaut l(sigma) - d(v_sigma(n-1),v_sigma(0)) donc
 T <= l(sigma) - d(v_sigma(n-1),v_sigma(0)) <= l(sigma)
*)




(* On veut générer les permutations sous forme d'un tableau par  ordre lexicographique. Quand on se donne une permutation
donnée par un tableau t, la suivante s'obtient de la manière suivante :
- on trouve le plus grand entier i tel que [|t.(i+1), ... t.(n-1)|] est décroissant ; si i = -1 on a fini c'est que l'on a 
la permutation [|n-1,n-2,...,1,0|]
- On inverse [|t.(i+1), ... t.(n-1)|]
- On permute t.(i) avec le premier élément de [|t.(i+1), ... t.(n-1)|] qui lui est supérieur.*)


let partie_decroissante t =
  let n = Array.length t in
  let k = ref (n - 1) in
  while !k >= 1 && t.(!k - 1) > t.(!k) do
    k := !k - 1
  done;
  !k - 1;;

(* determine i tel que [|t.(i),...,t.(n-1)|] soit décroissante *)




let echange t i j =
  let temp = t.(i) in
  t.(i) <- t.(j);
  t.(j) <- temp;;

let inverse t i =
  let n = Array.length t in
  let j = ref 1 in
  while i + !j <= n - !j do
    echange t (i + !j) (n - !j);
    incr j
  done;;

(* inverse   [|t.(i),...,t.(n-1)|] *)



let suivante t =
  let i = partie_decroissante t in
  inverse t i;
  let j = ref (i + 1) in
  while t.(i) > t.(!j) do
    incr j
  done;
  echange t i !j;;


let rec facto n = 
if n = 0 then 1 else n * facto (n-1);;

let voyageur_exhaustif points =
  let n = Array.length points in
  let stop = facto n in
  let sigmasol = ref (Array.make n 0) in 
  let sigma = Array.make n 0 in
  for j = 1 to (n - 1) do
    sigma.(j) <- j
  done;
  let longmin = ref (longueur_chemin points sigma) in
  for p = 1 to stop - 1 do
    suivante sigma;
    let l = longueur_chemin points sigma in
    if l < !longmin then (longmin := l; sigmasol := Array.copy sigma) 
  done;
  !longmin, !sigmasol;;

voyageur_exhaustif points;;
